package com.smardec.demo.regviewer;

import com.smardec.demo.tools.LibraryManager;
import com.smardec.j2native.*;

import javax.swing.*;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.event.TreeWillExpandListener;
import javax.swing.table.DefaultTableCellRenderer;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableCellRenderer;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeModel;
import javax.swing.tree.ExpandVetoException;
import javax.swing.tree.TreePath;
import java.awt.*;
import java.util.ArrayList;
import java.util.Iterator;

public class RegViewer extends JPanel{
    private JTree navTree;
    private DefaultTreeModel model;
    private DefaultMutableTreeNode root;
    private JTable dataTable;

    private boolean loaded;

    private Function regEnumKeyEx;
    private Function regEnumValue;
    private Function regOpenKey;
    private Function regCloseKey;

    public RegViewer(){
        super(new BorderLayout());
        root = new DefaultMutableTreeNode("My Computer");
        model = new DefaultTreeModel(root);
        navTree = new JTree(model);
        navTree.getSelectionModel().addTreeSelectionListener(new ViewerTreeSelectionListener());
        navTree.addTreeWillExpandListener(new ViewerTreeWillExpandListener());
        dataTable = createDataTable();

        JSplitPane splitPane = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
                                              new JScrollPane(navTree),
                                              new JScrollPane(dataTable));

        splitPane.setDividerLocation(250);
        add(splitPane, BorderLayout.CENTER);

        loaded = false;
    }

    public boolean initFunctions(){
        if(loaded){
            return true;
        }
        try{
            Library advapi32 = new Library("Advapi32");
            LibraryManager.getInstance().registerLibrary(advapi32);
            regEnumKeyEx = advapi32.getFunction("RegEnumKeyExA");
            regEnumValue = advapi32.getFunction("RegEnumValueA");
            regOpenKey = advapi32.getFunction("RegOpenKeyA");
            regCloseKey = advapi32.getFunction("RegCloseKey");

        } catch(LibraryNotFoundException e){
            e.printStackTrace();
            return false;
        } catch(FunctionNotFoundException e){
            e.printStackTrace();
            return false;
        }
        return true;
    }

    public void loadRegistry(){
        if(loaded){
            return;
        }
        RegNode node = new RegNode(new AnsiString("HKEY_CLASSES_ROOT"), new Pointer.Void(RWConstants.HKEY_CLASSES_ROOT));
        root.add(node);
        loadChildNodes(node, new Pointer.Void(RWConstants.HKEY_CLASSES_ROOT), 2);

        node = new RegNode(new AnsiString("HKEY_CURRENT_USER"), new Pointer.Void(RWConstants.HKEY_CURRENT_USER));
        root.add(node);
        loadChildNodes(node, new Pointer.Void(RWConstants.HKEY_CURRENT_USER), 2);

        node = new RegNode(new AnsiString("HKEY_LOCAL_MACHINE"), new Pointer.Void(RWConstants.HKEY_LOCAL_MACHINE));
        root.add(node);
        loadChildNodes(node, new Pointer.Void(RWConstants.HKEY_LOCAL_MACHINE), 2);

        node = new RegNode(new AnsiString("HKEY_USERS"), new Pointer.Void(RWConstants.HKEY_USERS));
        root.add(node);
        loadChildNodes(node, new Pointer.Void(RWConstants.HKEY_USERS), 2);

        node = new RegNode(new AnsiString("HKEY_CURRENT_CONFIG"), new Pointer.Void(RWConstants.HKEY_CURRENT_CONFIG));
        root.add(node);
        loadChildNodes(node, new Pointer.Void(RWConstants.HKEY_CURRENT_CONFIG), 2);

        loaded = true;
        navTree.expandRow(0);
    }

    private void loadChildNodes(RegNode parentNode, PointerArgument hKey, int levelDown){
        if(levelDown == 0){
            return;
        }

        UInt dwResult = new UInt();
        UInt dwIndex = new UInt();
        AnsiString name = new AnsiString("", 512);
        Pointer lpName = new Pointer(name);
        UInt cbName = new UInt();
        Pointer lpcbName = new Pointer(cbName);
        Pointer.Void lpReserved = new Pointer.Void();
        Pointer.Void lpClass = new Pointer.Void();
        Pointer.Void lpcbClass = new Pointer.Void();
        FILETIME ftLastWriteTime = new FILETIME();
        Pointer lpftLastWriteTime = new Pointer(ftLastWriteTime);
        while(true){
            cbName.setValue(512);
            regEnumKeyEx.invoke(dwResult, new Argument[]{hKey, dwIndex,
                                                         lpName, lpcbName, lpReserved, lpClass, lpcbClass, lpftLastWriteTime});
            if(dwResult.intValue() != RWConstants.ERROR_SUCCESS){
                break;
            }
            dwIndex.setValue(dwIndex.intValue() + 1);

            AnsiString childName = new AnsiString(name.toString().substring(0, cbName.intValue()));
            UInt lResult = new UInt();
            Pointer.Void hkResult = new Pointer.Void();
            Pointer phkResult = new Pointer(hkResult);
            regOpenKey.invoke(lResult, hKey, new Pointer(childName), phkResult);
            if(lResult.intValue() != RWConstants.ERROR_SUCCESS){
                continue;
            }

            RegNode childNode = new RegNode(childName);
            parentNode.add(childNode);
            loadChildNodes(childNode, hkResult, levelDown - 1);
            regCloseKey.invoke(null, hkResult);
        }
        ;
        if(levelDown > 1){
            parentNode.setTwoLevelLoaded(true);
        }
    }

    private int closeKey(PointerArgument hKey){
        UInt uResult = new UInt();
        regCloseKey.invoke(uResult, hKey);
        return uResult.intValue();
    }

    private int openKey(PointerArgument hKey, AnsiString childName, PointerArgument hKeyResult){
        UInt lResult = new UInt();
        regOpenKey.invoke(lResult, hKey, new Pointer(childName), new Pointer(hKeyResult));
        return lResult.intValue();
    }

    private UInt8[] dataMtx;

    private ArrayList retrieveData(PointerArgument hKey){
        ArrayList list = new ArrayList();
        UInt lResult = new UInt();

        UInt dwIndex = new UInt();
        AnsiString valueName = new AnsiString("", 512);
        Pointer lpValueName = new Pointer(valueName);
        UInt cValueName = new UInt();
        Pointer lpcValueName = new Pointer(cValueName);
        Pointer.Void lpReserved = new Pointer.Void();
        UInt type = new UInt();
        Pointer lpType = new Pointer(type);

        if(dataMtx == null){
            dataMtx = new UInt8[1024];
            for(int i = 0; i < dataMtx.length; i++){
                dataMtx[i] = new UInt8();
            }
        }
        ArrayArgument data = new ArrayArgument(dataMtx);

        Pointer lpData = new Pointer(data);
        UInt cData = new UInt();
        Pointer lpcbData = new Pointer(cData);

        while(true){
            cValueName.setValue(512);
            cData.setValue(1024);
            regEnumValue.invoke(lResult, new Argument[]{hKey, dwIndex, lpValueName, lpcValueName, lpReserved, lpType, lpData, lpcbData});
            if(lResult.intValue() != RWConstants.ERROR_SUCCESS){
                break;
            }
            dwIndex.setValue(dwIndex.intValue() + 1);

            String valueNameStr = valueName.toString().substring(0, cValueName.intValue());
            byte[] bData = new byte[cData.intValue()];
            System.arraycopy(data.toBytesValue(), 0, bData, 0, bData.length);

            list.add(new RegInfo(valueNameStr, type.intValue(), bData));
        }
        ;
        return list;
    }

    private ArrayList getInfoForLastNode(TreePath path){
        ArrayList info = null;
        if(path.getPathCount() > 1){
            PointerArgument hKey = ((RegNode)path.getPath()[1]).getHKey();
            if(path.getPathCount() > 2){
                info = getInfoForLastNode(hKey, path.getPath(), 2);
            } else{
                info = retrieveData(hKey);
            }
        }
        return info;
    }

    private ArrayList getInfoForLastNode(PointerArgument hKey, Object[] path, int index){
        int errorCode;
        ArrayList info;
        RegNode node = (RegNode)path[index];
        Pointer.Void hkResult = new Pointer.Void();
        if((errorCode = openKey(hKey, node.getName(), hkResult)) != RWConstants.ERROR_SUCCESS){
            System.err.println("Error #005 - error code: " + errorCode);
            return null;
        }
        index++;
        if(index < path.length){
            info = getInfoForLastNode(hkResult, path, index);
        } else{
            info = retrieveData(hkResult);
        }
        if((errorCode = closeKey(hkResult)) != RWConstants.ERROR_SUCCESS){
            System.err.println("Error #006 - error code: " + errorCode);
        }
        return info;
    }

    private JTable createDataTable(){
        DefaultTableModel tableModel = new DefaultTableModel(new String[]{"Name", "Type", "Data"}, 0){
            public boolean isCellEditable(int row, int column){
                return false;
            }
        };
        JTable table = new JTable(tableModel);
        table.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
        table.getColumnModel().getColumn(1).setCellRenderer(new TableCellRenderer(){
            JTextArea textArea = new JTextArea();

            public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column){
                String stringValue = value.toString();
                textArea.setText(stringValue);
                int prefHeight = textArea.getPreferredSize().height;
                if(prefHeight != table.getRowHeight(row)){
                    table.setRowHeight(row, prefHeight);
                }
                return textArea;
            }
        });
        table.setCellSelectionEnabled(false);
        table.getColumnModel().getColumn(0).setMinWidth(100);
        table.getColumnModel().getColumn(1).setMinWidth(100);
        table.getColumnModel().getColumn(1).setCellRenderer(new DefaultTableCellRenderer(){
            {
                setHorizontalAlignment(JLabel.CENTER);
            }
        });
        table.getColumnModel().getColumn(2).setMinWidth(120);
        table.getTableHeader().setReorderingAllowed(false);
        return table;
    }

    private class ViewerTreeSelectionListener implements TreeSelectionListener{
        public void valueChanged(TreeSelectionEvent e){
            if(!loaded){
                return;
            }
            TreePath path = e.getNewLeadSelectionPath();
            if(path == null){
                return;
            }
            Object lastPathComp = path.getLastPathComponent();
            DefaultTableModel tableModel = (DefaultTableModel)dataTable.getModel();
            tableModel.setRowCount(0);
            if(lastPathComp instanceof RegNode){
                StringBuffer stringBuffer = new StringBuffer();
                AnsiString ansiString = new AnsiString();
                Iterator iterator = getInfoForLastNode(path).iterator();
                while(iterator.hasNext()){
                    RegInfo regInfo = (RegInfo)iterator.next();
                    String type = "";
                    String data = "";
                    switch(regInfo.type){
                        case RWConstants.REG_NONE:
                            type = new String("REG_NONE");
                            for(int i = 0; i < regInfo.data.length; i++){
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_SZ:
                            type = new String("REG_SZ");
                            ansiString.fromBytesValue(regInfo.data, 0);
                            data = ansiString.toString();
                            break;
                        case RWConstants.REG_EXPAND_SZ:
                            type = new String("REG_EXPAND_SZ");
                            ansiString.fromBytesValue(regInfo.data, 0);
                            data = ansiString.toString();
                            break;
                        case RWConstants.REG_BINARY:
                            type = new String("REG_BINARY");
                            for(int i = 0; i < regInfo.data.length; i++){
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_DWORD:
                            type = new String("REG_DWORD");
                            UInt uInt = new UInt();
                            uInt.fromBytesValue(regInfo.data, 0);
                            data = "0x" + Integer.toHexString(uInt.intValue()) + "(" + uInt.intValue() + ")";
                            break;
                        case RWConstants.REG_DWORD_BIG_ENDIAN:
                            type = new String("REG_DWORD_BIG_ENDIAN");
                            Int64 int64 = new Int64();
                            int64.fromBytesValue(regInfo.data, 0);
                            data = "0x" + Long.toHexString(int64.longValue()) + "(" + int64.longValue() + ")";
                            break;
                        case RWConstants.REG_LINK:
                            type = new String("REG_LINK");
                            for(int i = 0; i < regInfo.data.length; i++){
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_MULTI_SZ:
                            type = new String("REG_MULTI_SZ");
                            int offset = 0;
                            while(offset < regInfo.data.length){
                                ansiString.fromBytesValue(regInfo.data, offset);
                                stringBuffer.append(ansiString.toString() + "\n");
                                offset += ansiString.getLength();
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_RESOURCE_LIST:
                            type = new String("REG_RESOURCE_LIST");
                            for(int i = 0; i < regInfo.data.length; i++){
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_FULL_RESOURCE_DESCRIPTOR:
                            type = new String("REG_FULL_RESOURCE_DESCRIPTOR");
                            for(int i = 0; i < regInfo.data.length; i++){
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                        case RWConstants.REG_RESOURCE_REQUIREMENTS_LIST:
                            type = new String("REG_RESOURCE_REQUIREMENTS_LIST");
                            for(int i = 0; i < regInfo.data.length; i++){
                                stringBuffer.append(Integer.toHexString((regInfo.data[i] & 0x00FF)) + " ");
                            }
                            data = stringBuffer.toString();
                            break;
                    }
                    if("".equals(regInfo.valueName)){
                        regInfo.valueName = "(Default)";
                    }
                    tableModel.addRow(new Object[]{regInfo.valueName, type, data});
                }
            }
        }
    }

    private class ViewerTreeWillExpandListener implements TreeWillExpandListener{
        public void treeWillCollapse(TreeExpansionEvent event) throws ExpandVetoException{
            if(event.getPath().getPathCount() == 1 || (!loaded)){
                throw new ExpandVetoException(null);
            }
        }

        public void treeWillExpand(TreeExpansionEvent event) throws ExpandVetoException{
            if(!loaded){
                throw new ExpandVetoException(null);
            }
            Object[] path = event.getPath().getPath();
            if((path.length > 2) && (!((RegNode)path[path.length - 1]).isTwoLevelLoaded())){
                loadSpecificChilds(((RegNode)path[1]).getHKey(), path, 2, 2);

            }
        }

        private void loadSpecificChilds(PointerArgument hKey, Object[] path, int index, int levelDown){
            int errorCode;
            RegNode node = (RegNode)path[index];
            Pointer.Void hkResult = new Pointer.Void();
            if((errorCode = openKey(hKey, node.getName(), hkResult)) != RWConstants.ERROR_SUCCESS){
                System.err.println("Error #001 - error code: " + errorCode);
                return;
            }
            index++;
            if(index < path.length){
                loadSpecificChilds(hkResult, path, index, levelDown);
            } else{
                for(int i = node.getChildCount() - 1; i >= 0; i--){
                    RegNode childNode = (RegNode)node.getChildAt(i);
                    Pointer.Void childHKey = new Pointer.Void();
                    if((errorCode = openKey(hkResult, childNode.getName(), childHKey)) != RWConstants.ERROR_SUCCESS){
                        System.err.println("Error #003 - error code: " + errorCode);
                        continue;
                    }

                    loadChildNodes(childNode, childHKey, levelDown);

                    if((errorCode = closeKey(childHKey)) != RWConstants.ERROR_SUCCESS){
                        System.err.println("Error #004 - error code: " + errorCode);
                    }
                }
                node.setTwoLevelLoaded(true);
            }
            if((errorCode = closeKey(hkResult)) != RWConstants.ERROR_SUCCESS){
                System.err.println("Error #002 - error code: " + errorCode);
            }
        }
    }

    private class RegInfo{
        public String valueName;
        public int type;
        public byte[] data;

        public RegInfo(String valueName, int type, byte[] data){
            this.valueName = valueName;
            this.type = type;
            this.data = data;
        }
    }

    private class FILETIME extends Structure{
        public UInt dwLowDateTime;
        public UInt dwHighDateTime;

        public FILETIME(){
            dwLowDateTime = new UInt();
            dwHighDateTime = new UInt();

            initialize(new Argument[]{dwLowDateTime, dwHighDateTime});
        }
    }
}
